package SevenZip;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;

import SevenZip.Compression.LZMA.Encoder;

public class LzmaBench {
    static final int kAdditionalSize = (6 << 20);
    static final int kCompressedAdditionalSize = (1 << 10);
    static final int kMaxLzmaPropSize = 10;
    
    static class CRandomGenerator {
        int A1;
        int A2;
        public CRandomGenerator() { Init(); }
        public void Init() { A1 = 362436069; A2 = 521288629; }
        public int GetRnd() {
            return
                    ((A1 = 36969 * (A1 & 0xffff) + (A1 >>> 16)) << 16) ^
                    ((A2 = 18000 * (A2 & 0xffff) + (A2 >>> 16)));
        }
    }
    
    static class CBitRandomGenerator {
        CRandomGenerator RG = new CRandomGenerator();
        int Value;
        int NumBits;
        public void Init() {
            Value = 0;
            NumBits = 0;
        }
        public int GetRnd(int numBits) {
            int result;
            if (NumBits > numBits) {
                result = Value & (((int)1 << numBits) - 1);
                Value >>>= numBits;
                NumBits -= numBits;
                return result;
            }
            numBits -= NumBits;
            result = (Value << numBits);
            Value = RG.GetRnd();
            result |= Value & (((int)1 << numBits) - 1);
            Value >>>= numBits;
            NumBits = 32 - numBits;
            return result;
        }
        public CBitRandomGenerator()  {}
    }
    
    static class CBenchRandomGenerator {
        CBitRandomGenerator RG = new CBitRandomGenerator();
        int Pos;
        public int BufferSize;
        public byte[] Buffer = null;
        public CBenchRandomGenerator() {  }
        public void Init() { RG.Init(); }
        public void Set(int bufferSize) {
            Buffer = new byte[bufferSize];
            Pos = 0;
            BufferSize = bufferSize;
        }
        int GetRndBit() { return RG.GetRnd(1); }
        
        int GetLogRandBits(int numBits) {
            int len = RG.GetRnd(numBits);
            return RG.GetRnd(len);
        }
        int GetOffset() {
            if (GetRndBit() == 0)
                return GetLogRandBits(4);
            return (GetLogRandBits(4) << 10) | RG.GetRnd(10);
        }
        int GetLen() {
            if (GetRndBit() == 0)
                return RG.GetRnd(2);
            if (GetRndBit() == 0)
                return 4 + RG.GetRnd(3);
            return 12 + RG.GetRnd(4);
        }
        public void Generate() {
            while (Pos < BufferSize) {
                if (GetRndBit() == 0 || Pos < 1)
                    Buffer[Pos++] = (byte)(RG.GetRnd(8));
                else {
                    int offset = GetOffset();
                    while (offset >= Pos)
                        offset >>>= 1;
                    offset += 1;
                    int len = 2 + GetLen();
                    for (int i = 0; i < len && Pos < BufferSize; i++, Pos++)
                        Buffer[Pos] = Buffer[Pos - offset];
                }
            }
        }
    }
    
    static class CProgressInfo implements SevenZip.ICodeProgress {
        public long ApprovedStart;
        public long InSize;
        long Time;
        public void Init() { InSize = 0; }
        public void SetProgress(long inSize, long outSize) {
            if (inSize >= ApprovedStart && InSize == 0) {
                Time = System.currentTimeMillis();
                InSize = inSize;
            }
        }
    }
    
    
    
    static final int kSubBits = 8;
    
    static int GetLogSize(int size) {
        for (int i = kSubBits; i < 32; i++)
            for (int j = 0; j < (1 << kSubBits); j++)
                if (size <= (((int)1) << i) + (j << (i - kSubBits)))
                    return (int)(i << kSubBits) + j;
        return (32 << kSubBits);
    }
    
    static long MyMultDiv64(long value, long elapsedTime) {
        long freq = 1000; // elapsedTime : millisecond;
        long elTime = elapsedTime;
        while (freq > 1000000) {
            freq >>>= 1;
            elTime >>>= 1;
        }
        if (elTime == 0)
            elTime = 1;
        return value * freq / elTime;
    }
    
    static long GetCompressRating(int dictionarySize, boolean isBT4,
            long elapsedTime, long size) {
        long numCommandsForOne;
        if (isBT4) {
            long t = GetLogSize(dictionarySize) - (19 << kSubBits);
            numCommandsForOne = 2000 + ((t * t * 68) >>> (2 * kSubBits));
        } else {
            long t = GetLogSize(dictionarySize) - (15 << kSubBits);
            numCommandsForOne = 1500 + ((t * t * 41) >>> (2 * kSubBits));
        }
        long numCommands = (long)(size) * numCommandsForOne;
        return MyMultDiv64(numCommands, elapsedTime);
    }
    
    static long GetDecompressRating(long elapsedTime, long outSize, long inSize) {
        long numCommands = inSize * 250 + outSize * 21;
        return MyMultDiv64(numCommands, elapsedTime);
    }
    
    static void PrintValue(long v) {
        String s = ""+v; // v.ToString();
        for (int i = 0; i + s.length() < 6; i++)
            System.out.print(" ");
        System.out.print(s);
    }
    
    static void PrintRating(long rating) {
        PrintValue(rating / 1000000);
        System.out.print(" MIPS");
    }
    
    static void PrintResults(
            int dictionarySize,
            boolean isBT4,
            long elapsedTime,
            long size,
            boolean decompressMode, long secondSize) {
        long speed = MyMultDiv64(size, elapsedTime);
        PrintValue(speed / 1024);
        System.out.print(" KB/s  ");
        long rating;
        if (decompressMode)
            rating = GetDecompressRating(elapsedTime, size, secondSize);
        else
            rating = GetCompressRating(dictionarySize, isBT4, elapsedTime, size);
        PrintRating(rating);
    }
    
    static class CRC {
        public static  int[] Table;
        
        static {
            Table = new int[256];
            int kPoly = (int)0xEDB88320;
            for (int i = 0; i < 256; i++) {
                int r = i;
                for (int j = 0; j < 8; j++)
                    if ((r & 1) != 0)
                        r = (r >>> 1) ^ kPoly;
                    else
                        r >>>= 1;
                Table[i] = r;
            }
        }
        
        int _value = -1; // 0xFFFFFFFF;
        
        public void Init() { _value = -1; /* 0xFFFFFFFF; */ }
        
        public void Update(byte[] data, int offset, int size) {
            for (int i = 0; i < size; i++)
                _value = Table[((_value) ^ data[offset + i])&0xFF] ^ (_value >>> 8);
        }
        
        public int GetDigest() { return _value ^ 0xFFFFFFFF; }
    }
    
    static class CrcOutStream extends java.io.OutputStream {
        public CRC CRC = new CRC();
        public void Init() { CRC.Init(); }
        public int GetDigest() { return CRC.GetDigest(); }
        
        public void write(byte[] b) {
            CRC.Update(b, 0, b.length);
        }
        public void write(byte[] b, int off, int len) {
            CRC.Update(b, off, len);
        }
        public void write(int b) {
            byte [] buf = new byte[1];
            buf[1]=(byte)b;
            CRC.Update(buf,0,1);
        }
    }
    
    public static int LzmaBenchmark(int numIterations, int dictionarySize, boolean isBT4) {
        try {
            if (numIterations <= 0)
                return 0;
            if (dictionarySize < (1 << 19) && isBT4 || dictionarySize < (1 << 15)) {
                System.out.println("\nError: dictionary size for benchmark must be >= 19 (512 KB)");
                return 1;
            }
            System.out.println("\n       Compressing                Decompressing\n\n");
            
            SevenZip.Compression.LZMA.Encoder encoder = new SevenZip.Compression.LZMA.Encoder();
            SevenZip.Compression.LZMA.Decoder decoder = new SevenZip.Compression.LZMA.Decoder();
            
            int[] propIDs =
            {
                CoderPropID.DictionarySize,
                        CoderPropID.MatchFinder
            };
            int[] properties =
            {
                dictionarySize,
                        isBT4 ? Encoder.EMatchFinderType_BT4: Encoder.EMatchFinderType_BT2
            };
            
            
            int kBufferSize = dictionarySize + kAdditionalSize;
            int kCompressedBufferSize = (kBufferSize / 2) + kCompressedAdditionalSize;
            
            encoder.SetCoderProperties(propIDs, properties);
            ByteArrayOutputStream propStream = new ByteArrayOutputStream();
            encoder.WriteCoderProperties(propStream);
            byte[] propArray = propStream.toByteArray();
            
            CBenchRandomGenerator rg = new CBenchRandomGenerator();
            
            rg.Init();
            rg.Set(kBufferSize);
            rg.Generate();
            
            CRC crc = new CRC();
            
            crc.Init();
            crc.Update(rg.Buffer, 0, rg.BufferSize);
            
            CProgressInfo progressInfo = new CProgressInfo();
            progressInfo.ApprovedStart = dictionarySize;
            
            long totalBenchSize = 0;
            long totalEncodeTime = 0;
            long totalDecodeTime = 0;
            long totalCompressedSize = 0;
            
            ByteArrayInputStream inStream = new ByteArrayInputStream(rg.Buffer, 0, (int)rg.BufferSize);
            ByteArrayOutputStream compressedStream = new ByteArrayOutputStream((int)kCompressedBufferSize);
            CrcOutStream crcOutStream = new CrcOutStream();
            for (int interation = 0; interation < numIterations; interation++) {
                progressInfo.Init();
                inStream.reset(); //.Seek(0, SeekOrigin.Begin);
                compressedStream.reset(); //.Seek(0, SeekOrigin.Begin);
                encoder.Code(inStream, compressedStream, progressInfo);
                long encodeTime = System.currentTimeMillis() - progressInfo.Time;
                
                long compressedSize = compressedStream.size();
                if (progressInfo.InSize == 0)
                    throw (new java.io.IOException("Internal ERROR 1282"));
                
                ByteArrayInputStream compressedStream2 = new ByteArrayInputStream(compressedStream.toByteArray());
                
                long decodeTime = 0;
                for (int j = 0; j < 2; j++) {
                    compressedStream2.reset(); // .Seek(0, SeekOrigin.Begin);
                    crcOutStream.Init();
                    
                    decoder.SetDecoderProperties(propArray);
                    long outSize = kBufferSize;
                    long startTime = System.currentTimeMillis();
                    decoder.Code(compressedStream2, crcOutStream, (long)outSize, null);
                    decodeTime = System.currentTimeMillis() - startTime;
                    if (crcOutStream.GetDigest() != crc.GetDigest())
                        throw new java.io.IOException("CRC Error");
                }
                long benchSize = kBufferSize - (long)progressInfo.InSize;
                PrintResults(dictionarySize, isBT4, encodeTime, benchSize, false, 0);
                System.out.print("     ");
                PrintResults(dictionarySize, isBT4, decodeTime, kBufferSize, true, compressedSize);
                System.out.println();
                totalBenchSize += benchSize;
                totalEncodeTime += encodeTime;
                totalDecodeTime += decodeTime;
                totalCompressedSize += (long)compressedSize;
            }
            System.out.println("---------------------------------------------------");
            PrintResults(dictionarySize, isBT4, totalEncodeTime, totalBenchSize, false, 0);
            System.out.print("     ");
            PrintResults(dictionarySize, isBT4, totalDecodeTime,kBufferSize * (long)numIterations, true, totalCompressedSize);
            System.out.println("    Average");
            return 0;
        } catch (IOException e) {
            System.out.println(e.getMessage());
            return 1;
        }
    }
}


